@@ -12,26 +12,30 @@ use PadWalker qw( peek_my );  # peek_our ); ???
 
 A "rule" function gets as argument a list:
 
-0 - a string to match
+0 - a string to match 
 1 - an optional "continuation"
 2 - a partially built match tree
 3 - a leaf pointer in the match tree
+4 - a grammar name
+5 - pos - TODO - change #0 ???
+6 - the whole string to match - TODO - unify with $_[0]
+7 - argument list - <subrule($x,$y)>
 
 it returns (or "yields"):
 
     undef - match failed
 
-or a hash containing:
+or a hash containing: - TODO - no longer needed
 
     state - a "continuation" or undef
     bool - an "assertion" (true/false)
     match - the "match" tree or undef
-    tail - the string tail or undef
+    tail - the string tail or undef - TODO - don't need this if 'pos' is used
     capture - the tree of captured things
     abort - the match was stopped by a { return } or a fail(),
            and it should not backtrack or whatever
     return - the block (sub-reference) that contains 'return'
-    label - the capture name (ident, number, or '')
+    label - the capture name (ident or '')
 
 Continuations are used for backtracking.
 
@@ -65,7 +69,7 @@ sub alternation {
         while ( defined $state ) {
             ### alternation string to match: "$tail - (node,state)=@$state"
             $match = 
-                $nodes->[ $state->[0] ]->( $tail, $state->[1], $_[2], $_[3]{match} );
+                $nodes->[ $state->[0] ]->( $tail, $state->[1], $_[2], $_[3]{match}, @_[4,5,6,7] );
             $match = $$match if ref($match) eq 'Pugs::Runtime::Match';
             ### match: $match
             if ( $match->{state} ) {
@@ -101,10 +105,10 @@ sub concat {
         $_[3] = { match => [] };
         while (1) {
             
-            $matches[0] = $nodes[0]->( $tail, $state[0], $_[2], $_[3]{match}[0] );
+            $matches[0] = $nodes[0]->( $tail, $state[0], $_[2], $_[3]{match}[0], @_[4,5,6,7] );
             $matches[0] = ${$matches[0]} if ref($matches[0]) eq 'Pugs::Runtime::Match';
             ### 1st match: $matches[0]
-            return $matches[0] 
+            return $_[3] = $matches[0] 
                 if $matches[0]{abort};
             if ( ! $matches[0]{bool} ) {
                 return unless defined $matches[0]{state};
@@ -115,15 +119,17 @@ sub concat {
             $_[3]{capture} = $_[3]{match}[0]{capture};
             #print "Matched concat 0, tree:", Dumper($_[2]);
 
-            $matches[1] = $nodes[1]->( $matches[0]{tail}, $state[1], $_[2], $_[3]{match}[1] );
+            $matches[1] = $nodes[1]->( $matches[0]{tail}, $state[1], $_[2], $_[3]{match}[1], @_[4,5,6,7] );
             $matches[1] = ${$matches[1]} if ref($matches[1]) eq 'Pugs::Runtime::Match';
             ### 2nd match: $matches[1]
+            if ( ! $matches[1]{abort} ) {
+                # die "not implemented";
+                # XXX - untested (use ':' to test this)
+                # XXX - what to do with $matches[0] ???
+                #return $_[3] = $matches[1];
+            #}
+            
             if ( ! $matches[1]{bool} ) {
-                
-                if ( $matches[1]{abort} ) {
-                    die "not implemented";
-                }
-
                 if ( ! defined( $matches[1]{state} ) ) {
                     return unless defined $matches[0]{state};
                     @state = ( $matches[0]{state}, 0 );
@@ -133,7 +139,7 @@ sub concat {
                 next;
             }
             
-
+            }
             #print "Matched concat 1, tree:", Dumper($_[2]) if defined $_[2];
 
             my $succ;
@@ -159,8 +165,11 @@ sub concat {
             $match2->{match} = \@matches;
             $match2->{state} = $succ;
             delete $match2->{label};
-            delete $matches[1]{abort};
+            #delete $matches[1]{abort};
             delete $matches[1]{return};
+
+            # print "concat: ",Dumper( $match2 );
+
             return $_[3] = $match2;
         }
     }
@@ -168,31 +177,43 @@ sub concat {
 
 sub constant { 
     my $const = shift;
+    my $lconst = length( $const );
+    no warnings qw( uninitialized );
     return sub {
-        ### matching constant:$_[0],$const
-        return if ! $_[0] || $_[0] !~ m/^(\Q$const\E)(.*)/s;
-        $_[3] = { 
-            bool => 1,
-            match => { constant => $1 }, 
-            tail => $2,
-        };
-        return { bool => 1,
-                 match => { constant => $1 }, 
-                 tail => $2,
-               }
+        #print "Pos: $_[5] - ", length($_[6])-length($_[0]), "\n";
+        #print "Runtime.constant: $const \n";
+        if ( $const eq substr( $_[0], $_[5], $lconst ) ) {
+            return $_[3] = { 
+                bool => 1,
+                match => $const, 
+                tail => substr( $_[0], $_[5]+$lconst ),
+            };
+        }
+        return $_[3] = { bool => 0 };
     }
 }
 
+sub perl5 {
+    my $rx = qr(^($_[0])(.*)$)s;
+    no warnings qw( uninitialized );
+    return sub {
+        if ( $_[0] =~ m/$rx/ ) {
+            return $_[3] = { 
+                bool  => 1,
+                match => $1,
+                tail  => $2,
+            };
+        }
+        return $_[3] = { bool => 0 };
+    };
+}
+
 sub null {
     return sub {
-        $_[3] = { 
+        return $_[3] = { 
             bool => 1,
             tail => $_[0],
         };
-        return { bool => 1,
-                 #match => undef,  #'null',
-                 tail => $_[0],
-               }
     }
 };
 
@@ -202,7 +223,7 @@ sub capture {
     my $node = shift;
     sub {
         $_[3] = { label => $label };
-        my $match = $node->( @_[0,1,2], $_[3]{match} );
+        my $match = $node->( @_[0,1,2], $_[3]{match}, @_[4,5,6,7] );
         $match = $$match if ref($match) eq 'Pugs::Runtime::Match';
         return unless $match->{bool};
         ## return if $match->{abort}; - maybe a { return }
@@ -244,6 +265,8 @@ sub try {
 sub abort { 
     my $op = shift;
     return sub {
+        #print __PACKAGE__ . "->abort\n";
+        #print caller;
         my $match = $op->( @_ );
         ### aborting match: $match
         $match->{abort} = 1;
@@ -251,6 +274,14 @@ sub abort {
     };
 };
 
+sub fail { 
+    return abort( 
+        sub {
+            return { bool => 0 };
+        } 
+    );
+};
+
 # experimental!
 sub negate { 
     my $op = shift;
@@ -259,7 +290,17 @@ sub negate {
         my $match = $op->( @_ );
         return if $match->{bool};
         return { bool => 1,
-                 #match => undef,  #'null',
+                 tail => $_[0],
+               }
+    };
+};
+
+sub before { 
+    my $op = shift;
+    return sub {
+        my $match = $op->( @_ );
+        #return $match unless $match->{bool};
+        return { bool => $match->{bool},
                  tail => $_[0],
                }
     };
@@ -286,23 +327,6 @@ sub wrap {
     }
 }
 
-sub perl5 {
-    my $rx = qr(^($_[0])(.*)$)s;
-    #print "rx: $rx\n";
-    return sub {
-        return unless defined $_[0];
-        if ( $_[0] =~ m/$rx/ ) {
-            return $_[3] = { 
-                bool  => 1,
-                match => $1,
-                tail  => $2,
-                capture => $1,
-            };
-        }
-        return;
-    };
-}
-
 
 # ------- higher-order ruleops
 
@@ -350,7 +374,7 @@ sub non_greedy_plus {
         # XXX - didn't work
         # my $match = $state->{op}->( $tail, $state->{state}, $flags ); 
 
-        my $match = $state->{op}->( $tail, undef );
+        my $match = $state->{op}->( $tail, undef, $_[2], $_[3]{match}, @_[4..6] );
         return unless $match->{bool};
         $match->{state} = {
             state => $match->{state},
@@ -402,9 +426,65 @@ sub get_variable {
         $idx++, next
           unless exists $pad->{$name};
 
-        return ${ $pad->{$name} };
+        #print "NAME $name $pad->{$name}\n";
+        return ${ $pad->{$name} } if $name =~ /^\$/;
+        return $pad->{$name};  # arrayref/hashref
     }
     die "Couldn't find '$name' in surrounding lexical scope.";
 }
 
+sub _preprocess_hash {
+    my $h = shift;
+    if ( ref($h) eq 'CODE') {
+        return sub {
+            $h->();
+            return { bool => 1, match => '', tail => $_[0] };
+        };
+    } 
+    if ( ref($h) eq 'Pugs::Compiler::Rule') {
+        #print "compiling subrule\n";
+        #return $h->code;
+        return sub { 
+            #print "into subrule - $_[0] - grammar $_[4]\n"; 
+            #print $h->code;
+            my $match = $h->match( $_[0], $_[4], { p => 1 } );
+            #print "match: ",$match->(),"\n";
+            return $_[3] = $$match;
+        };
+    }
+    # fail is number != 1 
+    if ( $h =~ /^(\d+)$/ ) {
+        return sub{} unless $1 == 1;
+        return sub{ { bool => 1, match => '', tail => $_[0] } };
+    }
+    # subrule
+    warn "uncompiled subrule: $h - not implemented";
+    return sub {};
+}
+
+# see commit #9783 for an alternate implementation
+sub hash {
+    my %hash = %{shift()};
+    #print "HASH: @{[ %hash ]}\n";
+    my @keys = sort {length $b <=> length $a } keys %hash;
+    #print "hash keys: @keys\n";
+    @keys = map {
+        concat(
+            constant( $_ ),
+            _preprocess_hash( $hash{$_} ),
+        )
+    } @keys;
+    return alternation( \@keys );
+}
+
+sub end_of_string {
+    return sub {
+        return $_[3] = { 
+            bool  => ($_[0] eq ''),
+            match => '',
+            tail  => $_[0],
+        };
+    };
+}
+
 1;
